/*
 * stack.h - header for stack.c
 *
 * Copyright 1993 Matthew Green
 * Copyright 1997 EPIC Software Labs
 * See the COPYRIGHT file for license information.
 */

#ifndef __stack_h__
#define __stack_h__

#define STACK_POP 	0
#define STACK_PUSH 	1
#define STACK_SWAP 	2
#define STACK_LIST 	3

#define STACK_DO_ALIAS	0x0001
#define STACK_DO_ASSIGN	0x0002

#endif /* __stack_h_ */
