/*
 * help.h: header for help.c
 *
 * Copyright 1994 Matthew Green
 * Copyright 1997 EPIC Software Labs
 * See the Copyright file for license information
 */

#ifndef __help_h__
#define __help_h__

	BUILT_IN_COMMAND(help);

#endif /* __help_h */
