/*
 * queue.h - header file for queue.c
 *
 * Copyright 1993 Jeremy Nelson
 * See the coypright file for more information
 */

#ifndef __queue_h__
#define __queue_h__

	BUILT_IN_COMMAND(queuecmd);

#endif
