/*
 * mail.h: header for mail.c
 *
 * Copyright 1990 Michael Sandrof
 * Copyright 1997 EPIC Software Labs
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef __mail_h__
#define __mail_h__

	char	*check_mail 		(void);
	int	check_mail_status 	(void *);

#endif /* _MAIL_H_ */
